import pandas as pd
from datetime import date
from thetadata import ThetaClient, OptionReqType, OptionRight


def get_last() -> pd.DataFrame:

    # Create a ThetaClient
    client = ThetaClient()

    # Connect to the Terminal
    with client.connect():

        # Request the most recent quote for AAPL
        quote = client.get_last_option(
            req=OptionReqType.QUOTE,
            root="AAPL",
            exp=date(2022, 9, 16),
            strike=140.00,
            right=OptionRight.CALL,
        )

    return quote


if __name__ == "__main__":
    quote = get_last()
    print(quote)



"""Package containing tests for the ThetaData Python API."""
import pytest
from thetadata import ThetaClient


@pytest.fixture
def tc():
    """Generate a ThetaClient connected to the Terminal."""
    client = ThetaClient(timeout=15, launch=False)
    with client.connect():
        yield client



import os
import sys
import urllib.request


def launch_terminal(username: str = None, passwd: str = None, use_bundle: bool = False):

    cmd = 'java -jar ThetaTerminal.jar'

    if use_bundle:
        jre = os.path.join(str(sys.path[1]), 'resources', 'corretto-11.0.16.1-win', 'bin', 'java')
        jar = os.path.join(str(sys.path[1]), 'resources', 'ThetaTerminal.jar')
        cmd = jre + ' -jar ' + jar

    if username is None or passwd is None:
        os.system(cmd)
    else:
        os.system(cmd + ' ' + username + ' ' + passwd)


def check_download(auto_update: bool):
    if os.path.exists('ThetaTerminal.jar') or auto_update:
        jar = urllib.request.urlopen("https://download-latest.thetadata.us")
        with open('ThetaTerminal.jar', 'wb') as output:
            output.write(jar.read())
            output.close()





"""Module that parses data from the Terminal."""
from __future__ import annotations
from typing import Optional

from tqdm import tqdm

from dataclasses import dataclass
import pandas as pd
import numpy as np
from pandas import DataFrame, Series
from .exceptions import ResponseError, NoData, ResponseParseError
from .enums import DataType, MessageType


@dataclass
class Header:
    """Represents the header returned on every Terminal call."""

    message_type: MessageType
    id: int
    latency: int
    error: int
    format_len: int
    size: int

    @classmethod
    def parse(cls, request: str, data: bytes) -> Header:
        """Parse binary header data into an object.

        :param request: the request that returned the header data
        :param data: raw header data, 20 bytes long
        :raises ResponseParseError: if parsing failed
        """
        try:
            return cls._parse(data)
        except Exception as e:
            raise ResponseParseError(
                f"Failed to parse header for request: {request}. Please send this error to support."
            ) from e

    @classmethod
    def _parse(cls, data: bytes) -> Header:
        """Parse binary header data into an object.

        :param data: raw header data, 20 bytes long
        """
        assert (
            len(data) == 20
        ), f"Cannot parse header with {len(data)} bytes. Expected 20 bytes."
        # avoid copying header data when slicing
        data = memoryview(data)
        """
        Header format:
            bytes | field
                2 | message type
                8 | id
                2 | latency
                2 | error
                1 | reserved / special
                1 | format length
                4 | size
        """
        parse_int = lambda d: int.from_bytes(d, "big")
        # parse
        msgtype = MessageType.from_code(parse_int(data[:2]))
        id = parse_int(data[2:10])
        latency = parse_int(data[10:12])
        error = parse_int(data[12:14])
        format_len = data[15]
        size = parse_int(data[16:20])
        return cls(
            message_type=msgtype,
            id=id,
            latency=latency,
            error=error,
            format_len=format_len,
            size=size,
        )


def _check_body_errors(header: Header, body_data: bytes):
    """Check for errors from the Terminal.

    :raises NoData: if the server does not contain data for the request.
    :raises ResponseError: if the header indicates an error, containing a
                           helpful error message.
    """
    if header.message_type == MessageType.ERROR:
        msg = body_data.decode("ascii")
        if "no data" in msg.lower():
            raise NoData(msg)
        else:
            raise ResponseError(msg)


# map price types to price multipliers
_pt_to_price_mul = [
    0,
    0.000000001,
    0.00000001,
    0.0000001,
    0.000001,
    0.00001,
    0.0001,
    0.001,
    0.01,
    0.1,
    1,
    10.0,
    100.0,
    1000.0,
    10000.0,
    100000.0,
    1000000.0,
    10000000.0,
    100000000.0,
    1000000000.0,
]

# Vectorized function that maps price types to price multipliers
_to_price_mul = np.vectorize(lambda pt: _pt_to_price_mul[pt], otypes=[float])


class TickBody:
    """Represents the body returned on Terminal calls that deal with ticks."""

    def __init__(self, format_tick: list[DataType], body_ticks: np.ndarray):
        assert isinstance(format_tick, list) and isinstance(
            body_ticks, np.ndarray
        ), "Cannot initialize body bc ticks is not a DataFrame"
        self.format_tick: list[DataType] = format_tick
        self.body_ticks: np.ndarray = body_ticks

    @classmethod
    def parse(cls, request: str, header: Header, data: bytearray) -> DataFrame:
        """Efficiently parse binary tick data.

        :param request: the request that returned the body data
        :param header: parsed header data
        :param data: the binary response body
        :return: a processed pandas dataframe
        :raises ResponseParseError: if parsing failed
        """
        assert isinstance(
            data, bytearray
        ), f"Expected data to be bytearray type. Got {type(data)}"
        _check_body_errors(header, data)
        try:
            tbody = cls._parse(header, data)
            df = tbody._to_dataframe()
            return df
        except Exception as e:
            raise ResponseParseError(
                f"Failed to parse body for request: {request}. Please send this error to support."
            ) from e

    @classmethod
    def _parse(cls, header: Header, data: bytearray) -> TickBody:
        assert (
            len(data) == header.size
        ), f"Cannot parse body with {len(data)} bytes. Expected {header.size} bytes."
        n_cols = header.format_len
        n_ticks = int(header.size / (header.format_len * 4))

        # parse format tick
        format: list[DataType] = []
        for ci in range(n_cols):
            int_ = int.from_bytes(data[ci * 4 : ci * 4 + 4], "big")
            format.append(DataType.from_code(int_))

        # parse the rest of the ticks
        # 4 byte integers w/ big endian order
        dtype = np.dtype("int32").newbyteorder(">")
        ticks = (
            np.frombuffer(data, dtype=dtype, offset=(header.format_len * 4))
            .reshape((n_ticks - 1), n_cols)
            .byteswap()  # force native byte order
            .newbyteorder()  # ^^
        )

        return cls(format_tick=format, body_ticks=ticks)

    def _to_dataframe(self) -> DataFrame:
        """Load this tick data into a pandas DataFrame and post process.

        Post-processing modifies the columns of data w/ various quality-of-life
        improvements.
        """
        # load ticks into DataFrame
        df = pd.DataFrame(
            self.body_ticks, columns=self.format_tick, copy=False
        )
        self._post_process(df)
        return df

    @classmethod
    def _post_process(cls, df: DataFrame) -> None:
        """Modify tick data w/ various quality-of-life improvements.

        :param df: The DataFrame to modify in-place.
        """
        # remove trailing null tick if it exists
        last_row = df.tail(1)
        zeroes = last_row.squeeze() == 0
        if zeroes.all():
            # print(f"Dropping {last_row=}")
            df.drop(last_row.index, inplace=True)

        if DataType.PRICE_TYPE in df.columns:
            # replace price type column with price multipliers

            df[DataType.PRICE_TYPE] = _to_price_mul(df[DataType.PRICE_TYPE])

            # multiply prices by price multipliers
            for col in df.columns:
                if col.is_price():
                    df[col] *= df[DataType.PRICE_TYPE]

            # remove price type column
            del df[DataType.PRICE_TYPE]

        # convert date ints to datetime
        if DataType.DATE in df.columns:
            df[DataType.DATE] = pd.to_datetime(
                df[DataType.DATE], format="%Y%m%d"
            )


class ListBody:
    """Represents the body returned on every Terminal call that have one DataType."""

    def __init__(self, lst: Series):
        assert isinstance(
            lst, Series
        ), "Cannot initialize body bc lst is not a Series"
        self.lst: Series = lst

    @classmethod
    def parse(
        cls, request: str, header: Header, data: bytes, dates: bool = False
    ) -> ListBody:
        """Parse binary body data into an object.

        :param request: the request that returned the header data
        :param header: parsed header data
        :param data: the binary response body
        :param dates: whether to parse the data as date objects
        :raises ResponseParseError: if parsing failed
        """
        _check_body_errors(header, data)
        try:
            return cls._parse(header, data, dates)
        except Exception as e:
            raise ResponseParseError(
                f"Failed to parse header for request: {request}. Please send this error to support."
            ) from e

    @classmethod
    def _parse(
        cls, header: Header, data: bytes, dates: bool = False
    ) -> ListBody:
        assert (
            len(data) == header.size
        ), f"Cannot parse body with {len(data)} bytes. Expected {header.size} bytes."

        lst = data.decode("ascii").split(",")
        lst = pd.Series(lst, copy=False)

        if dates:
            lst = pd.to_datetime(lst, format="%Y%m%d")

        return cls(lst=lst)



import pandas as pd
from thetadata import DataType
from .end_of_day import end_of_day


def main():
    # get the data from the previous example
    data: pd.DataFrame = end_of_day()

    # print all datatypes in the response
    print(f"{data.columns}")

    # get the first row in the DataFrame (our requested data)
    row = data.iloc[0]

    # print just the close price
    close_price = row[DataType.CLOSE]
    print(f"{close_price=}")


if __name__ == "__main__":
    main()



import pandas as pd
from thetadata import ThetaClient, SecType


def get_roots() -> pd.Series:
    # Create a ThetaClient
    client = ThetaClient()

    # Connect to the Terminal
    with client.connect():

        # List all root symbols for options
        roots = client.get_roots(sec=SecType.OPTION)

    return roots


if __name__ == "__main__":
    roots = get_roots()
    print(roots)



"""Module that contains custom exceptions."""
import enum
from typing import Any


class ResponseParseError(Exception):
    """Raised if the API failed to parse a Terminal response."""


class _EnumParseError(Exception):
    """Raised when a value cannot be parsed into an associated enum member."""

    def __init__(self, value: Any, enm: Any):
        """Create a new `EnumParseError`

        :param value: The value that cannot be parsed.
        :param enm: The enum.
        """
        assert issubclass(
            enm, enum.Enum
        ), "Cannot create an EnumParseError with a non-enum."
        msg = f"Value {value} cannot be parsed into a {enm.__name__}!"
        super().__init__(msg)


class ResponseError(Exception):
    """Raised if there is an error in the body of a response."""


class NoData(Exception):
    """Raised if no data is available for this request."""



"""Contains various tests for the ThetaClient class."""
import warnings

import pandas as pd
from pandas import DataFrame, Series
import pytest
import datetime
import thetadata
from thetadata import (
    ThetaClient,
    OptionReqType,
    OptionRight,
    DateRange,
    SecType,
    DataType, NoData,
)
from . import tc


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_end_of_day(tc: ThetaClient):
    """Test an EOD historical request."""
    res = tc.get_hist_option(
        req=OptionReqType.EOD,
        root="AAPL",
        exp=datetime.date(2022, 7, 15),
        strike=140,
        right=OptionRight.CALL,
        date_range=DateRange(
            datetime.date(2022, 7, 6), datetime.date(2022, 7, 15)
        ),
        progress_bar=True,
    )
    print(res)
    assert isinstance(res, DataFrame)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_hist_option_quotes_small(tc: ThetaClient):
    """Test a historical option request."""
    res = tc.get_hist_option(
        req=OptionReqType.QUOTE,
        root="AAPL",
        exp=datetime.date(2022, 7, 15),
        strike=140,
        right=OptionRight.CALL,
        date_range=DateRange(
            datetime.date(2022, 7, 6), datetime.date(2022, 7, 8)
        ),
        progress_bar=True,
    )
    print(res)
    print(res.columns)
    assert isinstance(res, DataFrame)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_hist_option_quotes_large(tc: ThetaClient):
    """Test a very large historical option request."""
    res = tc.get_hist_option(
        req=OptionReqType.QUOTE,
        root="AAPL",
        exp=datetime.date(2022, 7, 15),
        strike=140,
        right=OptionRight.CALL,
        date_range=DateRange(
            datetime.date(2022, 7, 4), datetime.date(2022, 7, 8)
        ),
        progress_bar=True,
    )
    print(res)
    assert isinstance(res, DataFrame)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_hist_option_trades(tc: ThetaClient):
    """Test a very large historical option request."""
    res = tc.get_hist_option(
        req=OptionReqType.TRADE,
        root="AAPL",
        exp=datetime.date(2022, 9, 16),
        strike=140,
        right=OptionRight.CALL,
        date_range=DateRange(datetime.date(2022, 9, 1), datetime.date(2022, 9, 16)),
        progress_bar=True,
    )
    print(res)
    assert isinstance(res, DataFrame)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_hist_option_open_interest(tc: ThetaClient):
    """Test a very large historical option request."""
    today = datetime.date.today()
    friday = today + datetime.timedelta((4 - today.weekday()) % 7)
    print("exp:        " + friday.__str__())
    print("root:       " + "AAPL")
    print("Date range: " + DateRange.from_days(7).__str__())

    try:
        res = tc.get_hist_option(
            req=OptionReqType.OPEN_INTEREST,
            root="AAPL",
            exp=datetime.date(2022, 11, 11),
            strike=140,
            right=OptionRight.CALL,
            date_range=DateRange(start=datetime.date(2022, 10, 29), end=datetime.date(2022, 11, 5)),
            progress_bar=True,
        )
    except NoData:
        today = datetime.date.today() + datetime.timedelta(days=7)
        friday = today + datetime.timedelta((4 - today.weekday()) % 7)
        res = tc.get_hist_option(
            req=OptionReqType.OPEN_INTEREST,
            root="AAPL",
            exp=friday,
            strike=140,
            right=OptionRight.CALL,
            date_range=DateRange.from_days(7),
            progress_bar=True,
        )
    print(res)
    assert isinstance(res, DataFrame)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_get_expirations(tc: ThetaClient):
    """Test an expirations listing request."""
    res = tc.get_expirations(root="AAPL")
    print(res)
    assert isinstance(res, Series)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_get_strikes_error(tc: ThetaClient):
    """Ensure that an invalid strike listing request raises."""
    with pytest.raises(thetadata.ResponseError) as e_info:
        res = tc.get_strikes(root="BDX", exp=datetime.date(2022, 6, 1))


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_get_strikes(tc: ThetaClient):
    """Test a strike listing request."""
    res = tc.get_strikes(
        root="AAPL",
        exp=datetime.date(2022, 7, 29),
    )
    print(res)
    assert isinstance(res, Series)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Ignore for now.")  # TODO: remove
def test_get_roots(tc: ThetaClient):
    """Test a root listing request."""
    res = tc.get_roots(sec=SecType.OPTION)
    print(res)
    assert isinstance(res, Series)
    assert len(res.index) > 0


@pytest.mark.skip(reason="Unable to retrieve last Price")  # TODO: remove
def test_get_last(tc: ThetaClient):
    """Test a get last option data request."""
    res = tc.get_last_option(
        req=OptionReqType.QUOTE,
        root="AAPL",
        exp=(datetime.datetime.now() + datetime.timedelta(days=4)).date(),
        strike=140,
        right=OptionRight.CALL,
    )
    print(res)
    assert isinstance(res, DataFrame)
    assert len(res.index) == 1


@pytest.mark.skip(reason="Cannot automate restart yet")
def test_kill_method(tc: ThetaClient):
    """Test killing the Terminal process by calling client.kill()"""
    tc.kill()



"""Module that contains various enums necessary to interface w the Terminal."""
from __future__ import annotations

import enum
from datetime import datetime, date, timedelta
from dataclasses import dataclass

from . import exceptions


@enum.unique
class DataType(enum.Enum):
    """Codes used in the format tick to ID the data in the body ticks."""

    DATE = (0, False)
    MS_OF_DAY = (1, False)
    CORRECTION = (2, False)
    PRICE_TYPE = (4, False)

    # QUOTES
    BID_SIZE = (101, False)
    BID_EXCHANGE = (102, False)
    BID = (103, True)
    BID_CONDITION = (104, False)
    ASK_SIZE = (105, False)
    ASK_EXCHANGE = (106, False)
    ASK = (107, True)
    ASK_CONDITION = (108, False)

    # PRICING
    MIDPOINT = (111, True)
    VWAP = (112, True)
    QWAP = (113, True)
    WAP = (114, True)

    # OPEN INTEREST
    OPEN_INTEREST = (121, True)

    # TRADES
    SEQUENCE = (131, False)
    SIZE = (132, False)
    CONDITION = (133, False)
    PRICE = (134, True)

    # VOLUME
    VOLUME = (141, False)
    COUNT = (142, False)

    # FIRST ORDER GREEKS
    THETA = (151, True)
    VEGA = (152, True)
    DELTA = (153, True)
    RHO = (154, True)
    EPSILON = (155, True)
    LAMBDA = (156, True)

    # SECOND ORDER GREEKS
    GAMMA = (161, True)
    VANNA = (162, True)
    CHARM = (163, True)
    VOMMA = (164, True)
    VETA = (165, True)
    VERA = (166, True)
    SOPDK = (167, True)

    # THIRD ORDER GREEKS
    SPEED = (171, True)
    ZOMMA = (172, True)
    COLOR = (173, True)
    ULTIMA = (174, True)

    # OTHER CALCS
    D1 = (181, True)
    D2 = (182, True)
    DUAL_DELTA = (183, True)
    DUAL_GAMMA = (184, True)

    # OHLC
    OPEN = (191, True)
    HIGH = (192, True)
    LOW = (193, True)
    CLOSE = (194, True)

    # IMPLIED VOLATILITY
    IMPLIED_VOL = (201, True)
    BID_IMPLIED_VOL = (202, True)
    ASK_IMPLIED_VOL = (203, True)
    UNDERLYING_PRICE = (204, True)

    # OTHER
    RATIO = (211, True)
    RATING = (212, True)

    # DIVIDEND
    EX_DATE = (221, False)
    RECORD_DATE = (222, False)
    PAYMENT_DATE = (223, False)
    ANN_DATE = (224, False)
    DIVIDEND_AMOUNT = (225, True)
    LESS_AMOUNT = (226, True)

    @classmethod
    def from_code(cls, code: int) -> DataType:
        """Create a DataType by its associated code.

        :raises EnumParseError: If the code does not match a DataType
        """
        for member in cls:
            if code == member.value[0]:
                return member
        raise exceptions._EnumParseError(code, cls)

    def code(self) -> int:
        """:return: The datatype code associated w this type."""
        return self.value[0]

    def is_price(self) -> bool:
        """Check if this DataType indicates a price."""
        return self.value[1]


@enum.unique
class MessageType(enum.Enum):
    """Codes used to ID types of requests/responses."""

    # Internal client communication
    CREDENTIALS = 0
    SESSION_TOKEN = 1
    INFO = 2
    METADATA = 3
    CONNECTED = 4

    # API communication
    PING = 100
    ERROR = 101
    DISCONNECTED = 102
    RECONNECTED = 103
    REQ_SYMS = 104
    SET_SYMS = 105
    CANT_CHANGE_SYMS = 106
    CHANGED_SYMS = 107
    KILL = 108

    # Client data
    HIST = 200
    ALL_EXPIRATIONS = 201
    ALL_STRIKES = 202
    HIST_END = 203
    LAST = 204
    ALL_ROOTS = 205
    LIST_END = 206

    # Experimental
    REQUEST_SERVER_LIST = 300
    REQUEST_OPTIMAL_SERVER = 301
    OPTIMAL_SERVER = 302
    PACKET = 303
    BAN_IP = 304
    POPULATION = 305

    @classmethod
    def from_code(cls, code: int) -> MessageType:
        """Create a MessageType by its associated code.

        :raises EnumParseError: If the code does not match a MessageType
        """
        for member in cls:
            if code == member.value:
                return member
        raise exceptions._EnumParseError(code, cls)


@enum.unique
class SecType(enum.Enum):
    """Security types."""

    OPTION = "OPTION"
    STOCK = "STOCK"
    FUTURE = "FUTURE"
    FORWARD = "FORWARD"
    SWAP = "SWAP"
    DEBT = "DEBT"
    CRYPTO = "CRYPTO"
    WARRANT = "WARRANT"


@enum.unique
class OptionRight(enum.Enum):
    """Option rights."""

    PUT = "P"
    CALL = "C"


@enum.unique
class OptionReqType(enum.Enum):
    """Option request type codes."""

    # FREE
    EOD = 1

    # VALUE
    QUOTE = 101
    VOLUME = 102
    OPEN_INTEREST = 103
    OHLC = 104

    # STANDARD
    TRADE = 201
    IMPLIED_VOLATILITY = 202
    GREEKS = 203
    LIQUIDITY = 204
    LIQUIDITY_PLUS = 205

    # PRO
    TRADE_GREEKS = 301
    GREEKS_SECOND_ORDER = 302
    GREEKS_THIRD_ORDER = 303
    ALT_CALCS = 304


@dataclass
class DateRange:
    """Represents an inclusive date range."""

    start: date
    end: date

    def __init__(self, start: date, end: date):
        """Create an inclusive date range between `start` and `end`."""
        assert isinstance(
            start, date
        ), f"Expected start to be a date object. Got {type(start)}."
        assert isinstance(
            start, date
        ), f"Expected end to be a date object. Got {type(end)}."
        self.start = start
        self.end = end
        assert (
            start <= end
        ), f"Start date {self.start} is greater than end date {self.end}!"

    @classmethod
    def from_days(cls, n: int) -> DateRange:
        """Create a date range that spans the past `n` days."""
        assert type(n) == int
        assert n >= 0, "n must be nonnegative"
        end = datetime.now().date()
        start = end - timedelta(days=n)
        return cls(start, end)



"""Module that contains Theta Client class."""
import os
import warnings
from decimal import Decimal
from threading import Thread
from time import sleep
from typing import Optional
from contextlib import contextmanager

import socket

from pandas import DataFrame
from tqdm import tqdm
import pandas as pd

from .enums import *
from .parsing import (
    Header,
    TickBody,
    ListBody,
)
from .terminal import check_download, launch_terminal

_NOT_CONNECTED_MSG = "You must esetablish a connection first."


def _format_strike(strike: float) -> int:
    """Round USD to the nearest tenth of a cent, acceptable by the terminal."""
    return round(strike * 1000)


def _format_date(dt: date) -> str:
    """Format a date obj into a string acceptable by the terminal."""
    return dt.strftime("%Y%m%d")


class ThetaClient:
    """A high-level, blocking client used to fetch market data."""

    def __init__(self, port: int = 11000, timeout: Optional[float] = 60, launch: bool = True,
                 username: str = "default", passwd: str = "default", auto_update: bool = True, use_bundle: bool = False):
        """Construct a client instance to interface with market data.

        :param port: The port number specified in the Theta Terminal config
        :param timeout: The max number of seconds to wait for a response before
            throwing a TimeoutError
        :param launch: If true, launches the terminal; False uses an existing external terminal instance.
        :param username: Theta Data username / email. When specified with a
            password, this class will launch the terminal in headless mode.
        :param passwd: Theta Data password. When specified with a username,
            this class will launch the terminal in headless mode.
        :param auto_update: If true, this class will automatically download the
            latest terminal version each time this class is instantiated. If
            false, the terminal will use the current jar terminal file. If none
            exists, it will download the latest version.
        """
        self.port: int = port
        self.timeout = timeout
        self._server: Optional[socket.socket] = None  # None while disconnected
        if os.system("java -version") != 0:
            warnings.warn("Java 11 is required to use this API. Terminating...")
            return

        if launch:
            if passwd.__contains__('%') or passwd.__contains__('?') or passwd.__contains__(' '):
                raise ConnectionError('Unable to connect! Your password contains illegal characters: %, ?, or  (space).'
                                      'Please change it at https://thetadata.net -> login -> forgot password.')
            if username == "default" or passwd == "default":
                warnings.warn("You are using the free version of Theta Data. You are currently limited to "
                              "20 requests / minute. A data subscription can be purchased at https://thetadata.net. "
                              "If you already have a Theta Data account, specify the username and passwd parameters.")

            check_download(auto_update)
            Thread(target=launch_terminal, args=[username, passwd, use_bundle]).start()
        else:
            print("You are not launching the terminal. This means you should have an external instance already running.")

    @contextmanager
    def connect(self):
        """Initiate a connection with the Theta Terminal on `localhost`.

        :raises ConnectionRefusedError: If the connection failed.
        :raises TimeoutError: If the timeout is set and has been reached.
        """

        try:
            for i in range(15):
                try:
                    self._server = socket.socket()
                    self._server.connect(("localhost", self.port))
                    self._server.settimeout(1)
                    break
                except ConnectionError:
                    if i == 14:
                        raise ConnectionError('Unable to connect to the local Theta Terminal process.'
                                              ' Try restarting your system.')
                    sleep(1)
            self._server.settimeout(self.timeout)
            yield
        finally:
            self._server.close()

    def _recv(self, n_bytes: int, progress_bar: bool = False) -> bytearray:
        """Wait for a response from the Terminal.
        :param n_bytes:       The number of bytes to receive.
        :param progress_bar:  Print a progress bar displaying download progress.
        :return:              A response from the Terminal.
        """
        assert self._server is not None, _NOT_CONNECTED_MSG

        # receive body data in parts
        MAX_PART_SIZE = 512  # 4 KiB recommended for most machines

        buffer = bytearray(n_bytes)
        bytes_downloaded = 0

        # tqdm disable=True is slow bc it still calls __new__, which takes nearly 4ms
        range_ = range(0, n_bytes, MAX_PART_SIZE)
        iterable = tqdm(range_, desc="Downloading") if progress_bar else range_

        for i in iterable:
            part_size = min(MAX_PART_SIZE, n_bytes - bytes_downloaded)
            bytes_downloaded += part_size
            part = self._server.recv(part_size)
            buffer[i : i + part_size] = part

        assert bytes_downloaded == n_bytes
        return buffer

    def kill(self) -> None:
        """Remotely kill the Terminal process.

        All subsequent requests will timeout until the Terminal is restarted.
        """
        # assert self._server is not None, _NOT_CONNECTED_MSG We should be able to kill the terminal no matter what
        kill_msg = f"MSG_CODE={MessageType.KILL.value}\n"
        self._server.sendall(kill_msg.encode("utf-8"))

    def get_hist_option(
        self,
        req: OptionReqType,
        root: str,
        exp: date,
        strike: float,
        right: OptionRight,
        date_range: DateRange,
        interval_size: int = 0,
        progress_bar: bool = False,
    ) -> pd.DataFrame:
        """
        Get historical option data.

        :param req:            The request type.
        :param root:           The root symbol.
        :param exp:            The expiration date. Must be after the start of `date_range`.
        :param strike:         The strike price in USD, rounded to 1/10th of a cent.
        :param right:          The right of an option.
        :param date_range:     The dates to fetch.
        :param interval_size:        The interval size in milliseconds. Applicable only to OHLC & QUOTE requests.
        :param progress_bar:   Print a progress bar displaying download progress.

        :return:               The requested data as a pandas DataFrame.
        :raises ResponseError: If the request failed.
        """
        assert self._server is not None, _NOT_CONNECTED_MSG
        # format data
        strike = _format_strike(strike)
        exp_fmt = _format_date(exp)
        start_fmt = _format_date(date_range.start)
        end_fmt = _format_date(date_range.end)

        # send request
        hist_msg = f"MSG_CODE={MessageType.HIST.value}&START_DATE={start_fmt}&END_DATE={end_fmt}&root={root}&exp={exp_fmt}&strike={strike}&right={right.value}&sec={SecType.OPTION.value}&req={req.value}&IVL={interval_size}\n"
        self._server.sendall(hist_msg.encode("utf-8"))

        # parse response header
        header_data = self._server.recv(20)
        header: Header = Header.parse(hist_msg, header_data)

        # parse response body
        body_data = self._recv(header.size, progress_bar=progress_bar)
        body: DataFrame = TickBody.parse(hist_msg, header, body_data)
        return body

    # LISTING DATA

    def get_expirations(self, root: str) -> pd.Series:
        """
        Get all option expirations.

        :param root: The root symbol.
        :return: All expirations that Theta Data provides data for (YYYYMMDD).
        :raises ResponseError: If the request failed.
        """
        assert self._server is not None, _NOT_CONNECTED_MSG
        out = f"MSG_CODE={MessageType.ALL_EXPIRATIONS.value}&root={root}\n"
        self._server.send(out.encode("utf-8"))
        header = Header.parse(out, self._server.recv(20))
        body = ListBody.parse(out, header, self._recv(header.size), dates=True)
        return body.lst

    def get_strikes(self, root: str, exp: date) -> pd.Series:
        """
        Get all option strike prices in US tenths of a cent.

        :param root: The root symbol.
        :param exp: The expiration date.
        :return: The strike prices on the expiration.
        :raises ResponseError: If the request failed.
        """
        assert self._server is not None, _NOT_CONNECTED_MSG
        assert isinstance(exp, date)
        exp_fmt = _format_date(exp)
        out = f"MSG_CODE={MessageType.ALL_STRIKES.value}&root={root}&exp={exp_fmt}\n"
        self._server.send(out.encode("utf-8"))
        header = Header.parse(out, self._server.recv(20))
        body = ListBody.parse(out, header, self._recv(header.size)).lst
        div = Decimal(1000)

        s = pd.Series([], dtype='float64')
        c = 0
        for i in body:
            s[c] = Decimal(i) / div
            c += 1

        return s

    def get_roots(self, sec: SecType) -> pd.Series:
        """
        Get all roots for a certain security type.

        :param sec: The type of security.
        :return: All root symbols for the security type.
        :raises ResponseError: If the request failed.
        """
        assert self._server is not None, _NOT_CONNECTED_MSG
        out = f"MSG_CODE={MessageType.ALL_ROOTS.value}&sec={sec.value}\n"
        self._server.send(out.encode("utf-8"))
        header = Header.parse(out, self._server.recv(20))
        body = ListBody.parse(out, header, self._recv(header.size))
        return body.lst

    # LIVE DATA

    def get_last_option(
        self,
        req: OptionReqType,
        root: str,
        exp: date,
        strike: float,
        right: OptionRight,
    ) -> pd.DataFrame:
        """
        Get the most recent option data.

        :param req:            The request type.
        :param root:           The root symbol.
        :param exp:            The expiration date.
        :param strike:         The strike price in USD, rounded to 1/10th of a cent.
        :param right:          The right of an option.

        :return:               The requested data as a pandas DataFrame.
        :raises ResponseError: If the request failed.
        """
        assert self._server is not None, _NOT_CONNECTED_MSG
        # format data
        strike = _format_strike(strike)
        exp_fmt = _format_date(exp)

        # send request
        hist_msg = f"MSG_CODE={MessageType.LAST.value}&root={root}&exp={exp_fmt}&strike={strike}&right={right.value}&sec={SecType.OPTION.value}&req={req.value}\n"
        self._server.sendall(hist_msg.encode("utf-8"))

        # parse response
        header: Header = Header.parse(hist_msg, self._server.recv(20))
        body: DataFrame = TickBody.parse(
            hist_msg, header, self._recv(header.size)
        )

        return body

    def get_req(
        self,
        req: str,
    ) -> pd.DataFrame:
        assert self._server is not None, _NOT_CONNECTED_MSG
        # send request
        req = req + "\n"
        self._server.sendall(req.encode("utf-8"))

        # parse response header
        header_data = self._server.recv(20)
        header: Header = Header.parse(req, header_data)

        # parse response body
        body_data = self._recv(header.size, progress_bar=False)
        body: DataFrame = TickBody.parse(req, header, body_data)
        return body




"""Contains tests for the tutorial section of the documentation."""
import pytest
import docs
from docs.get_last import get_last
from docs.list_roots import get_roots
from docs.end_of_day import end_of_day
from docs import manipulate_df, manipulate_series

@pytest.mark.skip(reason="No data for contract")
def test_docs_end_of_day():
    """Test the end of day blog snippet."""
    end_of_day()
    manipulate_df.main()

@pytest.mark.skip(reason="No data for contract")
def test_docs_list_roots():
    """Test the list roots tutorial."""
    get_roots()
    manipulate_series.main()


@pytest.mark.skip(
    reason="Live data currently only works during trading hours."
)  # TODO: remove
def test_docs_get_last():
    """Test the live data tutorial."""
    get_last()



import pandas as pd
from .list_roots import get_roots


def main():
    # get the result from the previous example
    roots: pd.Series = get_roots()

    # Check if a symbol is in the Series
    contains_apple = "AAPL" in roots.values

    # Count symbols
    count = len(roots)

    # Filter symbols to those that start with A
    filtered = [s for s in roots if s.startswith("A")]
    for symbol in filtered:
        # do something
        pass

    print(f"{contains_apple=}")
    print(f"{count=}")
    print(f"{len(filtered)=}")


if __name__ == "__main__":
    main()



import pandas as pd
from datetime import date
from thetadata import ThetaClient, OptionReqType, OptionRight, DateRange


def end_of_day() -> pd.DataFrame:
    """Request end-of-day data between 7/18/2022 and 7/22/2022."""
    # Create a ThetaClient
    client = ThetaClient()

    # Connect to the Terminal
    with client.connect():

        # Make the request
        data = client.get_hist_option(
            req=OptionReqType.EOD,
            root="AAPL",
            exp=date(2022, 8, 12),
            strike=140,
            right=OptionRight.CALL,
            date_range=DateRange(date(2022, 7, 18), date(2022, 7, 22)),
        )

    return data


if __name__ == "__main__":
    data = end_of_day()
    print(data)
